"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var semver = require("semver");
function isAllSame(array) {
    if (array.length <= 1) {
        return true;
    }
    var first = array[0];
    for (var i = 1, len = array.length; i < len; i++) {
        if (first !== array[i]) {
            return false;
        }
    }
    return true;
}
var re_dependency = /^(@?[^@]+)@(.*)$/;
function parseDependency(dependencyStr) {
    var matched = dependencyStr.match(re_dependency);
    return {
        name: matched[1],
        semver: matched[2]
    };
}
function parsePackageMap(lockData) {
    var packageMap = new Map();
    var dependencies = lockData.object;
    Object.keys(dependencies).forEach(function (expectedDependency) {
        var resolved = dependencies[expectedDependency];
        var depInfo = parseDependency(expectedDependency);
        var tgtMap = packageMap.get(depInfo.name) || new Map();
        tgtMap.set(depInfo.semver, resolved);
        packageMap.set(depInfo.name, tgtMap);
    });
    return packageMap;
}
function serializePackageMap(packageMap) {
    var objects = {};
    packageMap.forEach(function (semverMap, name) {
        semverMap.forEach(function (resolved, semverStr) {
            objects[name + "@" + semverStr] = resolved;
        });
    });
    return {
        type: 'success',
        object: objects
    };
}
function isDuplicate(packageSemverMap) {
    return !isAllSame(Array.from(packageSemverMap.values()).map(function (item) { return item.version; }));
}
function findBestDedupeResultIndex(dedupeResults) {
    var maxItemInfo;
    var isBetterThan = function (item, targetItem) {
        if (item.ranges.size > targetItem.ranges.size) {
            return true;
        }
        else if (item.ranges.size === targetItem.ranges.size) {
            return semver.gt(item.target.version, targetItem.target.version);
        }
        return false;
    };
    dedupeResults.forEach(function (item, index) {
        if (!maxItemInfo || isBetterThan(item, maxItemInfo.item)) {
            maxItemInfo = { item: item, index: index };
        }
    });
    return maxItemInfo ? maxItemInfo.index : -1;
}
function findDedupeVersions(packageSemverMap) {
    var dedupeVersions = new Map();
    packageSemverMap.forEach(function (lockTarget) {
        var version = lockTarget.version;
        var dedupeSource = dedupeVersions.get(version) || { target: lockTarget, satisfiedRanges: new Set() };
        dedupeVersions.set(version, dedupeSource);
        packageSemverMap.forEach(function (t, range) {
            if (semver.satisfies(version, range)) {
                dedupeSource.satisfiedRanges.add(range);
            }
        });
    });
    var availableDedupeResults = [];
    dedupeVersions.forEach(function (dedupeSource) { return availableDedupeResults.push({ ranges: dedupeSource.satisfiedRanges, target: dedupeSource.target, }); });
    var sortedDedupeResults = [];
    var _loop_1 = function () {
        var bestDedupeResultIndex = findBestDedupeResultIndex(availableDedupeResults);
        if (bestDedupeResultIndex < 0) {
            return "break";
        }
        var bestDedupeResult = availableDedupeResults[bestDedupeResultIndex];
        availableDedupeResults.splice(bestDedupeResultIndex, 1);
        sortedDedupeResults.push(bestDedupeResult);
        var handledRanges = bestDedupeResult.ranges;
        availableDedupeResults = availableDedupeResults.filter(function (item) {
            handledRanges.forEach(function (r) { return item.ranges.delete(r); });
            return item.ranges.size > 1;
        });
    };
    while (availableDedupeResults.length) {
        var state_1 = _loop_1();
        if (state_1 === "break")
            break;
    }
    return sortedDedupeResults
        .filter(function (item) {
        return item.ranges.size > 1 && Array.from(item.ranges).some(function (range) { return packageSemverMap.get(range).version !== item.target.version; });
    });
}
function dedupe(lockData, options) {
    if (options === void 0) { options = {}; }
    var packageMap = parsePackageMap(lockData);
    var dedupeResults = {
        resolved: [],
        partialResolved: [],
        unresolved: []
    };
    var shouldDedupeFor = function (packageName) {
        return !options.targetPackages || options.targetPackages.some(function (pkgFilter) {
            return typeof pkgFilter === 'string' ? pkgFilter === packageName : pkgFilter.test(packageName);
        });
    };
    packageMap.forEach(function (pkgSemverMap, packageName) {
        if (shouldDedupeFor(packageName) && isDuplicate(pkgSemverMap)) {
            var dedupeVersions = findDedupeVersions(pkgSemverMap);
            dedupeVersions.forEach(function (dedupeItem) {
                dedupeItem.ranges.forEach(function (range) {
                    pkgSemverMap.set(range, dedupeItem.target);
                });
            });
            if (dedupeVersions.length === 0) {
                dedupeResults.unresolved.push(packageName);
            }
            else if (dedupeVersions.length === 1 && dedupeVersions[0].ranges.size === pkgSemverMap.size) {
                dedupeResults.resolved.push(packageName);
            }
            else {
                dedupeResults.partialResolved.push(packageName);
            }
        }
    });
    return {
        lockData: serializePackageMap(packageMap),
        resolved: dedupeResults.resolved,
        partialResolved: dedupeResults.partialResolved,
        unresolved: dedupeResults.unresolved
    };
}
exports.dedupe = dedupe;
